/* 
    L64781  -  DVB-T COFDM demodulator used in Technotrend DVB-T cards

    Copyright (C) 1999 Holger Waechtler
                       for Convergence Integrated Media GmbH

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/    

#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <linux/i2c.h>

#include "dvb.h"
#include "dvb_demodulator.h"


static struct i2c_driver dvbt_driver;
static struct i2c_client client_template;




static
int write_reg(struct i2c_client *client, u8 reg, u8 data)
{
	int ret;
	u8 msg[2] = { reg, data };

	if (reg < 0x02 || (reg > 0x31 && reg != 0x3e)) {
		printk ("L64781: try to write readonly register 0x%02x !", reg);
		return -1;
	}

	if ((ret = i2c_master_send (client, msg, 2)) != 2)
		printk ("L64781: write_reg error !\n");

	mdelay(10);

	return ret;
}


static
u8 read_reg(struct i2c_client *client, u8 reg)
{
        struct i2c_adapter *adap=client->adapter;
        struct i2c_msg msgs[2];
        u8 data;

        msgs[0].addr = client->addr;
        msgs[0].len = 1;
        msgs[0].flags = 0;
        msgs[0].buf = &reg;
        msgs[1].addr = client->addr;
        msgs[1].len = 1;
        msgs[1].flags = I2C_M_RD;
	msgs[1].buf = &data;

	i2c_transfer(adap, msgs, 2);

        return data;
}


static
void L64781_dump_regs (struct i2c_client *client)
{
	int i;

	printk ("%s:\n", __FUNCTION__);
	printk ("static u8 L64781_regdump [] = {");
	
	for (i=0; i<=0x3e; i++) {
		if ((i & 0x7) == 0)
			printk ("\n        ");

		printk ("0x%02x", read_reg (client, i));

		if (i < 0x3e)
			printk (", ");
	}

	printk ("\n};\n");
}


static u8 init_regdump [0x4b] = {
	0x81, 0x00, 0x00, 0x00, 0x02, 0x01, 0x01, 0x8e,
	0x92, 0x80, 0xff, 0x81, 0x84, 0x8c, 0x3b, 0x6c,
	0x00, 0x09, 0x2c, 0xfd, 0x44, 0x03, 0x93, 0x18,
	0x04, 0x93, 0xa1, 0x39, 0x8e, 0x37, 0x49, 0x12,
	0x36, 0x36, 0xbe, 0x3f, 0x25, 0x2c, 0xd1, 0xfc,
	0x09, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x00,
	0x00, 0x00, 0x7f, 0x30, 0x00, 0x00, 0x00, 0x00,
	0x00, 0xc0, 0x01, 0xb0, 0xcb, 0xc0, 0x00, 0x00,
	0x28, 0x28, 0x00, 0x00, 0x00, 0x00, 0x01, 0x10,
	0x00, 0x00, 0x07
};


static
void apply_tps (struct i2c_client *client)
{
printk("L64781: apply_tps\n");
	write_reg (client, 0x2a, 0x01);
	write_reg (client, 0x2a, 0x00);
	write_reg (client, 0x2a, 0x06);
	write_reg (client, 0x2a, 0x00);
}

#if 0
static int qam_tab [] = { 2, 4, 0, 6 }; /* QPSK, QAM_16, QAM_64 */
static int bw_tab [] = { 8, 7, 6 };     /* 8Mhz, 7MHz, 6MHz */
static float guard_tab [] = { 1.0/32.0, 1.0/16.0, 1.0/8.0, 1.0/4.0 };
#endif

static
void apply_frontend_param (struct i2c_client *client, FrontendParameters *param)
{
	OFDMParameters *p = &param->u.ofdm;
/*	float spi_bias;

	spi_bias = 1512 * vcr_tab[p->HP_CodeRate];
        spi_bias *= qam_tab[p->Constellation] * 1024;
        spi_bias *= bw_tab[p->bandWidth];
        spi_bias /= 1 + guard_tab[p->guardInterval];
	spi_bias /= 8 * (1 + e);
*/
/* FIXME: uncomment for ZDF */
	write_reg (client, 0x11, 0x24);

/*
	printk ("mode == %u\n", p->TransmissionMode);
	printk ("guard == %u\n", p->guardInterval);
	printk ("hp_coderate == %u\n", p->HP_CodeRate);
	printk ("lp_coderate == %u\n", p->LP_CodeRate);
	printk ("guard == %u\n", p->guardInterval);
	printk ("L64781: write 0x04: 0x%02x\n", (u8)
		(p->TransmissionMode << 2) | p->guardInterval);
	printk ("L64781: write 0x05: 0x%02x\n", (u8)
		((p->LP_CodeRate - FEC_1_2) << 3) | (p->HP_CodeRate - FEC_1_2));
	printk ("L64781: write 0x06: 0x%02x\n", (u8)
		(p->Hierarchy << 2) | p->Constellation);
*/
	write_reg (client, 0x04, (p->TransmissionMode << 2) | p->guardInterval);
	write_reg (client, 0x05, ((p->LP_CodeRate - FEC_1_2) << 3) |
				  (p->HP_CodeRate - FEC_1_2));
	write_reg (client, 0x06, (p->HierarchyInformation << 2) | p->Constellation);
}


static
int init (struct i2c_client *client)
{
	u8 tmp_addr = client_template.addr;
	u8 msg = 0x06;

printk("L64781: init chip\n");

	client_template.addr = 0x0;
	i2c_master_send(&client_template, &msg, 1);
	client_template.addr = tmp_addr;

	L64781_dump_regs (client);

	{
	   u8 reg;

	   for (reg=0x00; reg<=0x4e; reg++)
		   write_reg (client, reg, init_regdump[reg]);
	}

	apply_tps (client);

	L64781_dump_regs (client);

	return 0;
}



static
int attach_adapter(struct i2c_adapter *adap)
{
        struct i2c_client *client;
	u8 tmp_addr = client_template.addr;
	u8 msg = 0x06;

	client_template.adapter=adap;

	if (i2c_master_send(&client_template,NULL,0) == 0) {
		printk("L64781: probably an EEPROM @ 0x55, skipping ...\n");
                return -1;      /*  no L64781, proabably an EEPROM ... */
	}

	client_template.addr = 0x0;

	i2c_master_send(&client_template, &msg, 1);
	client_template.addr = tmp_addr;

	if (i2c_master_send(&client_template,NULL,0)) {
		printk("L64781: no L64781 found ...\n");
                return -1;      /*  no L64781 found ... */
	}

        client_template.adapter=adap;
        
        if (!(client = kmalloc(sizeof(struct i2c_client), GFP_KERNEL)))
                return -ENOMEM;

        memcpy(client, &client_template, sizeof(struct i2c_client));
        
        printk("L64781: attaching L64781 at 0x%02x\n", (client->addr)<<1);

        i2c_attach_client(client);
        
	MOD_INC_USE_COUNT;
        
        return 0;
}


static
int detach_client(struct i2c_client *client)
{
        printk("L64781: detach_client\n");
        i2c_detach_client(client);
        kfree(client->data);
        kfree(client);
	MOD_DEC_USE_COUNT;
        return 0;
}


static 
void L64781_dump_stats (struct i2c_client *client)
{
	unsigned int gain = 0, afc_offset = 0, tim_offset = 0, dtr_freq = 0, ddfs_offset = 0, spi_gain = 0, spi_bias = 0;

	gain |= read_reg(client, 0x0e);
	if (gain & 0x80)
		gain |= 0xffffff00;

	afc_offset |= read_reg(client, 0x08);
	afc_offset |= read_reg(client, 0x09) << 8;
	afc_offset |= read_reg(client, 0x0a) << 16;
	if (afc_offset & 0x800000)
		afc_offset |= 0xff000000;
	
	tim_offset |= read_reg(client, 0x12);
	tim_offset |= read_reg(client, 0x13) << 8;

	dtr_freq |= read_reg(client, 0x16);
	dtr_freq |= read_reg(client, 0x17) << 8;
	dtr_freq |= read_reg(client, 0x18) << 16;
	if (dtr_freq & 0x800000)
		dtr_freq |= 0xff000000;

	ddfs_offset |= read_reg(client, 0x22);
	ddfs_offset |= read_reg(client, 0x23) << 8;

	spi_gain |= read_reg(client, 0x19) >> 3;
	spi_gain |= read_reg(client, 0x20) & 0x1f;

	spi_bias |= read_reg(client, 0x1b);
	spi_bias |= read_reg(client, 0x1c) << 8;
	spi_bias |= (read_reg(client, 0x1d) & 0x7f) << 16;

	printk ("L64781: gain == 0x%08x        (%i)\n", gain, gain);
	printk ("L64781: dtr_freq == 0x%08x    (%i)\n", dtr_freq, dtr_freq);
  	printk ("L64781: afc offset == 0x%08x  (%i)\n", afc_offset, afc_offset);
  	printk ("L64781: TIM offset == 0x%08x  (%i)\n", tim_offset, tim_offset);
	printk ("L64781: ddfs_offset == 0x%08x (%i)\n", ddfs_offset, ddfs_offset);
	printk ("L64781: spi_gain == 0x%08x    (%i)\n", spi_gain, spi_gain);
	printk ("L64781: spi_bias == 0x%08x    (%i)\n", spi_bias, spi_bias);
}


static 
void L64781_dump_TPS (struct i2c_client *client)
{
	u8 reg0x04 = read_reg (client, 0x04);
	u8 reg0x05 = read_reg (client, 0x05);
	u8 reg0x06 = read_reg (client, 0x06);

	printk ("xmit mode %u | ", (reg0x04 >> 2) & 0x3);
	printk ("guard interval %u | ", reg0x04 & 0x3);
	printk ("LP code rate %u | ", (reg0x05 >> 3) & 0x7);
	printk ("HP code rate %u | ", reg0x05 & 0x7);
	printk ("frame no %u | ", (reg0x06 >> 5) & 0x3);
	printk ("hierarchy info %u | ", (reg0x06 >> 2) & 0x7);
	printk ("constellation %u\n", reg0x06 & 0x3);
}

static 
int dvb_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
        switch (cmd) {
	case DVB_INIT_FRONTEND:
	{
		printk ("%s: DVB_INIT_FRONTEND\n", __FUNCTION__);
        	init(client);
                break;
	}
        case DVB_RESET:
        {
		printk ("%s: DVB_RESET\n", __FUNCTION__);
                break;
	}
	case DVB_SET_FRONTEND:
        {
                struct frontend *front = (struct frontend *)arg;
                
		printk ("%s: DVB_SET_FRONTEND\n", __FUNCTION__);

		printk("front->freq == %i\n", front->param.Frequency);

		apply_frontend_param (client, &front->param);
		apply_tps (client);

		break;
        }
        case DVB_GET_FRONTEND:
        {
                struct frontend *front = (struct frontend *)arg;

/*		front->afc =(read_reg(client, 0x08) << 16)
		          | (read_reg(client, 0x09) << 8)
		          | (read_reg(client, 0x0a) << 0);
*/		front->agc = read_reg(client,0x0c) & 0x7;
		front->fsync = read_reg(client,0x32);
		front->sync = front->fsync;
		front->nest = 0;
		front->vber = read_reg(client, 0x39)
			   | (read_reg(client, 0x3a) << 8);
//L64781_dump_regs (client);
//printk ("sync == 0x%02x\n", front->sync);
//L64781_dump_stats (client);
//if (read_reg(client, 0x0) & 1) L64781_dump_TPS (client);
		break;
        }
        default:
		//printk ("%s: unknown command !!!\n", __FUNCTION__);
                return -1;
        };
        
        return 0;
} 

void inc_use (struct i2c_client *client)
{
#ifdef MODULE
        MOD_INC_USE_COUNT;
#endif
}

void dec_use (struct i2c_client *client)
{
#ifdef MODULE
        MOD_DEC_USE_COUNT;
#endif
}

static
struct i2c_driver dvbt_driver = {
	name: 		"L64781 DVB demodulator",
	id:		I2C_DRIVERID_L64781,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	attach_adapter,
	detach_client:	detach_client,
	command:	dvb_command,
	inc_use:	inc_use,
	dec_use:	dec_use
};

static
struct i2c_client client_template = {
        name:    "L64781",
        id:      I2C_DRIVERID_L64781,
        flags:   0,
        addr:    0x55,
        adapter: NULL,
        driver:  &dvbt_driver
};


static __init
int init_l64781 (void) {
        int res;
        
        if ((res = i2c_add_driver(&dvbt_driver))) 
        {
                printk("L64781: i2c driver registration failed\n");
                return res;
        }

        printk("L64781: init done\n");
        return 0;
}

static __exit
void exit_l64781 (void)
{
        int res;

        if ((res = i2c_del_driver(&dvbt_driver))) 
        {
                printk("L64781: Driver deregistration failed, "
                       "module not removed.\n");
        }
        printk("L64781: cleanup\n");
}

module_init(init_l64781);
module_exit(exit_l64781);

